@@ -4,12 +4,13 @@ import (
 	"encoding/json"
 	"os"
 	"path/filepath"
+	"slices"
 	"testing"
 
+	compressiontypes "github.com/containers/image/v5/pkg/compression/types"
 	"github.com/containers/image/v5/types"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
-	"golang.org/x/exp/slices"
 )
 
 // assertJSONEqualsFixture tests that jsonBytes is structurally equal to fixture,
@@ -40,3 +41,13 @@ func layerInfosWithCryptoOperation(input []types.BlobInfo, op types.LayerCrypto)
 	}
 	return res
 }
+
+// layerInfosWithCompressionEdits returns a copy of input where CompressionOperation and CompressionAlgorithm is set to op and algo
+func layerInfosWithCompressionEdits(input []types.BlobInfo, op types.LayerCompression, algo *compressiontypes.Algorithm) []types.BlobInfo {
+	res := slices.Clone(input)
+	for i := range res {
+		res[i].CompressionOperation = op
+		res[i].CompressionAlgorithm = algo
+	}
+	return res
+}
