@@ -3,15 +3,15 @@ package signature
 import (
 	"context"
 	"fmt"
-	"os"
 	"testing"
 
-	"github.com/containers/image/v4/docker"
-	"github.com/containers/image/v4/docker/policyconfiguration"
-	"github.com/containers/image/v4/docker/reference"
-	"github.com/containers/image/v4/internal/testing/mocks"
-	"github.com/containers/image/v4/transports"
-	"github.com/containers/image/v4/types"
+	"github.com/containers/image/v5/docker"
+	"github.com/containers/image/v5/docker/policyconfiguration"
+	"github.com/containers/image/v5/docker/reference"
+	"github.com/containers/image/v5/internal/private"
+	"github.com/containers/image/v5/internal/testing/mocks"
+	"github.com/containers/image/v5/transports"
+	"github.com/containers/image/v5/types"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
 )
@@ -26,7 +26,10 @@ func TestPolicyRequirementError(t *testing.T) {
 func TestPolicyContextChangeState(t *testing.T) {
 	pc, err := NewPolicyContext(&Policy{Default: PolicyRequirements{NewPRReject()}})
 	require.NoError(t, err)
-	defer pc.Destroy()
+	defer func() {
+		err := pc.Destroy()
+		require.NoError(t, err)
+	}()
 
 	require.Equal(t, pcReady, pc.state)
 	err = pc.changeState(pcReady, pcInUse)
@@ -67,6 +70,7 @@ func TestPolicyContextNewDestroy(t *testing.T) {
 // pcImageReferenceMock is a mock of types.ImageReference which returns itself in DockerReference
 // and handles PolicyConfigurationIdentity and PolicyConfigurationReference consistently.
 type pcImageReferenceMock struct {
+	mocks.ForbiddenImageReference
 	transportName string
 	ref           reference.Named
 }
@@ -94,18 +98,6 @@ func (ref pcImageReferenceMock) PolicyConfigurationNamespaces() []string {
 	}
 	return policyconfiguration.DockerReferenceNamespaces(ref.ref)
 }
-func (ref pcImageReferenceMock) NewImage(ctx context.Context, sys *types.SystemContext) (types.ImageCloser, error) {
-	panic("unexpected call to a mock function")
-}
-func (ref pcImageReferenceMock) NewImageSource(ctx context.Context, sys *types.SystemContext) (types.ImageSource, error) {
-	panic("unexpected call to a mock function")
-}
-func (ref pcImageReferenceMock) NewImageDestination(ctx context.Context, sys *types.SystemContext) (types.ImageDestination, error) {
-	panic("unexpected call to a mock function")
-}
-func (ref pcImageReferenceMock) DeleteImage(ctx context.Context, sys *types.SystemContext) error {
-	panic("unexpected call to a mock function")
-}
 
 func TestPolicyContextRequirementsForImageRefNotRegisteredTransport(t *testing.T) {
 	transports.Delete("docker")
@@ -132,7 +124,7 @@ func TestPolicyContextRequirementsForImageRefNotRegisteredTransport(t *testing.T
 	require.NoError(t, err)
 	ref, err := reference.ParseNormalizedNamed("registry.access.redhat.com/rhel7:latest")
 	require.NoError(t, err)
-	reqs := pc.requirementsForImageRef(pcImageReferenceMock{"docker", ref})
+	reqs := pc.requirementsForImageRef(pcImageReferenceMock{transportName: "docker", ref: ref})
 	assert.True(t, &(reqs[0]) == &(pr[0]))
 	assert.True(t, len(reqs) == len(pr))
 
@@ -152,6 +144,8 @@ func TestPolicyContextRequirementsForImageRef(t *testing.T) {
 		{"docker", ""},
 		{"docker", "unmatched"},
 		{"docker", "deep.com"},
+		{"docker", "*.very.deep.com"},
+		{"docker", "*.deep.com"},
 		{"docker", "deep.com/n1"},
 		{"docker", "deep.com/n1/n2"},
 		{"docker", "deep.com/n1/n2/n3"},
@@ -178,10 +172,13 @@ func TestPolicyContextRequirementsForImageRef(t *testing.T) {
 		{"docker", "deep.com/n1/notn2/n3/repo:tag2", "docker", "deep.com/n1"},
 		// Host name match
 		{"docker", "deep.com/notn1/n2/n3/repo:tag2", "docker", "deep.com"},
+		// Sub domain match
+		{"docker", "very.deep.com/n1/n2/n3/repo:tag2", "docker", "*.deep.com"},
+		{"docker", "not.very.deep.com/n1/n2/n3/repo:tag2", "docker", "*.very.deep.com"},
 		// Default
-		{"docker", "this.doesnt/match:anything", "docker", ""},
+		{"docker", "this.does-not/match:anything", "docker", ""},
 		// No match within a matched transport which doesn't have a "" scope
-		{"atomic", "this.doesnt/match:anything", "", ""},
+		{"atomic", "this.does-not/match:anything", "", ""},
 		// No configuration available for this transport at all
 		{"dir", "what/ever", "", ""}, // "what/ever" is not a valid scope for the real "dir" transport, but we only need it to be a valid reference.Named.
 	} {
@@ -196,7 +193,7 @@ func TestPolicyContextRequirementsForImageRef(t *testing.T) {
 
 		ref, err := reference.ParseNormalizedNamed(c.input)
 		require.NoError(t, err)
-		reqs := pc.requirementsForImageRef(pcImageReferenceMock{c.inputTransport, ref})
+		reqs := pc.requirementsForImageRef(pcImageReferenceMock{transportName: c.inputTransport, ref: ref})
 		comment := fmt.Sprintf("case %s:%s: %#v", c.inputTransport, c.input, reqs[0])
 		// Do not use assert.Equal, which would do a deep contents comparison; we want to compare
 		// the pointers. Also, == does not work on slices; so test that the slices start at the
@@ -206,12 +203,11 @@ func TestPolicyContextRequirementsForImageRef(t *testing.T) {
 	}
 }
 
-// pcImageMock returns a types.UnparsedImage for a directory, claiming a specified dockerReference and implementing PolicyConfigurationIdentity/PolicyConfigurationNamespaces.
-// The caller must call the returned close callback when done.
-func pcImageMock(t *testing.T, dir, dockerReference string) (types.UnparsedImage, func() error) {
+// pcImageMock returns a private.UnparsedImage for a directory, claiming a specified dockerReference and implementing PolicyConfigurationIdentity/PolicyConfigurationNamespaces.
+func pcImageMock(t *testing.T, dir, dockerReference string) private.UnparsedImage {
 	ref, err := reference.ParseNormalizedNamed(dockerReference)
 	require.NoError(t, err)
-	return dirImageMockWithRef(t, dir, pcImageReferenceMock{"docker", ref})
+	return dirImageMockWithRef(t, dir, pcImageReferenceMock{transportName: "docker", ref: ref})
 }
 
 func TestPolicyContextGetSignaturesWithAcceptedAuthor(t *testing.T) {
@@ -254,88 +250,79 @@ func TestPolicyContextGetSignaturesWithAcceptedAuthor(t *testing.T) {
 		},
 	})
 	require.NoError(t, err)
-	defer pc.Destroy()
+	defer func() {
+		err := pc.Destroy()
+		require.NoError(t, err)
+	}()
 
 	// Success
-	img, closer := pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
-	defer closer()
+	img := pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
 	sigs, err := pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Equal(t, []*Signature{expectedSig}, sigs)
 
 	// Two signatures
 	// FIXME? Use really different signatures for this?
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid-2", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid-2", "testing/manifest:latest")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Equal(t, []*Signature{expectedSig, expectedSig}, sigs)
 
 	// No signatures
-	img, closer = pcImageMock(t, "fixtures/dir-img-unsigned", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-unsigned", "testing/manifest:latest")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Empty(t, sigs)
 
 	// Only invalid signatures
-	img, closer = pcImageMock(t, "fixtures/dir-img-modified-manifest", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-modified-manifest", "testing/manifest:latest")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Empty(t, sigs)
 
 	// 1 invalid, 1 valid signature (in this order)
-	img, closer = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:latest")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Equal(t, []*Signature{expectedSig}, sigs)
 
 	// Two sarAccepted results for one signature
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:twoAccepts")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:twoAccepts")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Equal(t, []*Signature{expectedSig}, sigs)
 
 	// sarAccepted+sarRejected for a signature
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:acceptReject")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:acceptReject")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Empty(t, sigs)
 
 	// sarAccepted+sarUnknown for a signature
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:acceptUnknown")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:acceptUnknown")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Equal(t, []*Signature{expectedSig}, sigs)
 
 	// sarRejected+sarUnknown for a signature
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:rejectUnknown")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:rejectUnknown")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Empty(t, sigs)
 
 	// sarUnknown only
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:unknown")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:unknown")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Empty(t, sigs)
 
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:unknown2")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:unknown2")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Empty(t, sigs)
 
 	// Empty list of requirements (invalid)
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:invalidEmptyRequirements")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:invalidEmptyRequirements")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	require.NoError(t, err)
 	assert.Empty(t, sigs)
@@ -347,8 +334,7 @@ func TestPolicyContextGetSignaturesWithAcceptedAuthor(t *testing.T) {
 	require.NoError(t, err)
 	err = destroyedPC.Destroy()
 	require.NoError(t, err)
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
 	sigs, err = destroyedPC.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	assert.Error(t, err)
 	assert.Nil(t, sigs)
@@ -358,9 +344,7 @@ func TestPolicyContextGetSignaturesWithAcceptedAuthor(t *testing.T) {
 
 	// Error reading signatures.
 	invalidSigDir := createInvalidSigDir(t)
-	defer os.RemoveAll(invalidSigDir)
-	img, closer = pcImageMock(t, invalidSigDir, "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, invalidSigDir, "testing/manifest:latest")
 	sigs, err = pc.GetSignaturesWithAcceptedAuthor(context.Background(), img)
 	assert.Error(t, err)
 	assert.Nil(t, sigs)
@@ -393,66 +377,59 @@ func TestPolicyContextIsRunningImageAllowed(t *testing.T) {
 		},
 	})
 	require.NoError(t, err)
-	defer pc.Destroy()
+	defer func() {
+		err := pc.Destroy()
+		require.NoError(t, err)
+	}()
 
 	// Success
-	img, closer := pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
-	defer closer()
+	img := pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
 	res, err := pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningAllowed(t, res, err)
 
 	// Two signatures
 	// FIXME? Use really different signatures for this?
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid-2", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid-2", "testing/manifest:latest")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningAllowed(t, res, err)
 
 	// No signatures
-	img, closer = pcImageMock(t, "fixtures/dir-img-unsigned", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-unsigned", "testing/manifest:latest")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningRejectedPolicyRequirement(t, res, err)
 
 	// Only invalid signatures
-	img, closer = pcImageMock(t, "fixtures/dir-img-modified-manifest", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-modified-manifest", "testing/manifest:latest")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningRejectedPolicyRequirement(t, res, err)
 
 	// 1 invalid, 1 valid signature (in this order)
-	img, closer = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:latest")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningAllowed(t, res, err)
 
 	// Two allowed results
-	img, closer = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:twoAllows")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:twoAllows")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningAllowed(t, res, err)
 
 	// Allow + deny results
-	img, closer = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:allowDeny")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:allowDeny")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningRejectedPolicyRequirement(t, res, err)
 
 	// prReject works
-	img, closer = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:reject")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:reject")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningRejectedPolicyRequirement(t, res, err)
 
 	// prInsecureAcceptAnything works
-	img, closer = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:acceptAnything")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-mixed", "testing/manifest:acceptAnything")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningAllowed(t, res, err)
 
 	// Empty list of requirements (invalid)
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:invalidEmptyRequirements")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:invalidEmptyRequirements")
 	res, err = pc.IsRunningImageAllowed(context.Background(), img)
 	assertRunningRejectedPolicyRequirement(t, res, err)
 
@@ -461,8 +438,7 @@ func TestPolicyContextIsRunningImageAllowed(t *testing.T) {
 	require.NoError(t, err)
 	err = destroyedPC.Destroy()
 	require.NoError(t, err)
-	img, closer = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
-	defer closer()
+	img = pcImageMock(t, "fixtures/dir-img-valid", "testing/manifest:latest")
 	res, err = destroyedPC.IsRunningImageAllowed(context.Background(), img)
 	assertRunningRejected(t, res, err)
 	// Not testing the pcInUse->pcReady transition, that would require custom PolicyRequirement
@@ -487,7 +463,7 @@ func assertSARRejected(t *testing.T, sar signatureAcceptanceResult, parsedSig *S
 	assert.Error(t, err)
 }
 
-// assertSARRejectedPolicyRequiremnt verifies that isSignatureAuthorAccepted returns a consistent sarRejected resul,
+// assertSARRejectedPolicyRequirement verifies that isSignatureAuthorAccepted returns a consistent sarRejected result,
 // and that the returned error is a PolicyRequirementError..
 func assertSARRejectedPolicyRequirement(t *testing.T, sar signatureAcceptanceResult, parsedSig *Signature, err error) {
 	assertSARRejected(t, sar, parsedSig, err)
